import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

/**
   This program demonstrates a generic insertion sort
   method that uses a Comparator<T> object to sort
   a List<T> of items.
*/
public class GenericInsertionSort_ver4
{
   public static void main(String[] args)
   {
      List<Triple> list1 = new ArrayList<>(
                                 Arrays.asList(
                                   new Triple(2, 3, 2),
                                   new Triple(4, 4, 1),
                                   new Triple(2, 4, 0),
                                   new Triple(5, 1, 1),
                                   new Triple(0, 6, 3),
                                   new Triple(2, 2, 2),
                                   new Triple(4, 1, 4),
                                   new Triple(5, 0, 4) ) );
      System.out.println(list1);
      sort(list1, new TripleComparator());
      System.out.println(list1);
      System.out.println();


      List<Country> list2 = new ArrayList<>(
                                  Arrays.asList(
                                    new Country("Freedonia", 3, 1),
                                    new Country("Shire", 5000, 50),
                                    new Country("Ambrosia", 10, 20),
                                    new Country("Utopia", 1000, 1000000),
                                    new Country("Neverland", 0, 0),
                                    new Country("Elbonia", 10, 20) ) );
      System.out.println(list2);
      sort(list2, new CountryComparator());
      System.out.println(list2);
      System.out.println();


      List<Pair> list3 = new ArrayList<>(
                                 Arrays.asList(
                                   new Pair(3, 1),
                                   new Pair(10, 20),
                                   new Pair(3, 0),
                                   new Pair(0, 0) ) );
      System.out.println(list3);
      sort(list3, new PairComparator());
      System.out.println(list3);
   }



   /**
      Sort a List<T> of objects using a Comparator<T> object
      and the iterative insertion sort algorithm.

      @param list  the List<T> to sort
   */
   public static <T> void sort(List<T> list, Comparator<T> cmp)
   {
      for (int i = 1; i < list.size(); ++i)
      {
         T next = list.remove(i);
         // Insert next into the list where it belongs.
         int j = i;
         while (j > 0 && cmp.compare(list.get(j - 1), next) > 0)
         {
            j--;
         }
         // Insert the element.
         list.add(j, next);
      }
   }
}



class TripleComparator implements Comparator<Triple>
{
   @Override public int compare(Triple first, Triple second)
   {
      if (first.x != second.x)
      {
         return first.x - second.x;
      }
      else if (first.y != second.y)
      {
         return first.y - second.y;
      }
      else
      {
         return first.z - second.z;
      }
   }
}


class CountryComparator implements Comparator<Country>
{
   @Override public int compare(Country first, Country second)
   {
      if (! first.name.equals(second.name) )
      {
         return first.name.compareTo(second.name);
      }
      else if (first.population != second.population)
      {
         return first.population - second.population;
      }
      else
      {
         return first.area - second.area;
      }
   }
}


class PairComparator implements Comparator<Pair>
{
   @Override public int compare(Pair first, Pair second)
   {
      if (first.x != second.x)
      {
         return first.x - second.x;
      }
      else
      {
         return first.y - second.y;
      }
   }
}
